@@ -23,6 +23,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <wchar.h>
+#include <wctype.h>
 #ifdef HAVE_NCURSESW_CURSES_H
   #include <ncursesw/curses.h>
 #else
@@ -462,8 +463,6 @@ void deletefiles_ncurses(file_t *files, char *logfile)
     exit(1);
   }
 
-  register_sigint_handler();
-
   curfile = files;
   while (curfile)
   {
